var REMOVED_ENTITY = '@@REMOVED_ENTITY@@';

var ArraySchema = /*#__PURE__*/function () {
  function ArraySchema(shape, options) {
    if (options === void 0) {
      options = {};
    }

    this.shape = shape;

    this.merge = options.merge || function (_, bb) {
      return bb;
    };

    this.create = options.create || function (aa) {
      return aa;
    };
  }

  var _proto = ArraySchema.prototype;

  _proto.normalize = function normalize(data, entities) {
    var _this = this;

    if (entities === void 0) {
      entities = {};
    }

    var schemas = {};
    var result = data.map(function (item) {
      var _this$shape$normalize = _this.shape.normalize(item, entities),
          result = _this$shape$normalize.result,
          childSchemas = _this$shape$normalize.schemas;

      Object.assign(schemas, childSchemas);
      return result;
    });
    return {
      entities: entities,
      schemas: schemas,
      result: this.create(result)
    };
  };

  _proto.denormalize = function denormalize(denormalizeState, path) {
    var _this2 = this;

    if (path === void 0) {
      path = [];
    }

    var result = denormalizeState.result,
        entities = denormalizeState.entities;
    if (result == null) return result;
    return result.map(function (item) {
      return _this2.shape.denormalize({
        result: item,
        entities: entities
      }, path);
    }).filter(function (ii) {
      return ii !== REMOVED_ENTITY;
    });
  };

  return ArraySchema;
}();

function UndefinedIdError(name, value) {
  return new Error(name + ".id() returned " + String(value) + ". Entities need some sort of id so they can be stored and retrieved. You should check the schema to see what is going on.");
}

var EntitySchema = /*#__PURE__*/function () {
  function EntitySchema(name, options) {
    var _options$shape;

    if (options === void 0) {
      options = {};
    }

    this.name = name;
    this.merge = options.merge;
    this.shape = (_options$shape = options.shape) != null ? _options$shape : null;

    this.id = options.id || function (data) {
      return data == null ? void 0 : data.id;
    };
  }

  var _proto = EntitySchema.prototype;

  _proto.normalize = function normalize(data, entities) {
    if (entities === void 0) {
      entities = {};
    }

    var shape = this.shape,
        name = this.name;
    var id = this.id(data);
    var previousEntity = null;
    var schemas = {};
    var result;
    if (id == null) throw UndefinedIdError(name, id);
    id = id.toString();
    entities[name] = entities[name] || {}; // only recurse if we have a defined shape

    if (shape) {
      var _ = shape.normalize(data, entities);

      result = _.result;
      schemas = _.schemas;
      previousEntity = entities[name][id];
    } else {
      result = data;
    } // list this schema as one that has been used


    schemas[name] = this; // Store the entity

    entities[name][id] = previousEntity && shape ? (this.merge || shape.merge)(previousEntity, result) : result;
    return {
      entities: entities,
      schemas: schemas,
      result: id
    };
  };

  _proto.denormalize = function denormalize(denormalizeState, path) {
    var _entities$name;

    if (path === void 0) {
      path = [];
    }

    var result = denormalizeState.result,
        entities = denormalizeState.entities;
    var shape = this.shape,
        name = this.name;
    var entity = (_entities$name = entities[name]) == null ? void 0 : _entities$name[result];
    if (entity == null || entity === REMOVED_ENTITY || shape === null) return entity != null ? entity : null;
    return shape.denormalize({
      result: entity,
      entities: entities
    }, path);
  };

  return EntitySchema;
}();

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

var ObjectSchema = /*#__PURE__*/function () {
  function ObjectSchema(relations, options) {
    if (options === void 0) {
      options = {};
    }

    this.relations = relations;

    this.create = options.create || function (item) {
      return _extends({}, item);
    };

    this.merge = options.merge || function (previous, next) {
      return _extends({}, previous, next);
    };
  }
  /**
   * ObjectSchema.normalize
   */


  var _proto = ObjectSchema.prototype;

  _proto.normalize = function normalize(data, entities) {
    if (entities === void 0) {
      entities = {};
    }

    var relations = this.relations;
    var dataMap = data;
    var schemas = {};
    var result = Object.keys(relations).reduce(function (result, key) {
      var value = dataMap[key];
      var schema = relations[key];
      if (!schema) throw new Error(String(key) + " was not defined in shape");

      if (value) {
        var _extends2;

        var _schema$normalize = schema.normalize(value, entities),
            childResult = _schema$normalize.result,
            childSchemas = _schema$normalize.schemas;

        Object.assign(schemas, childSchemas);
        result = _extends({}, result, (_extends2 = {}, _extends2[key] = childResult, _extends2));
      }

      return result;
    }, dataMap);
    return {
      entities: entities,
      schemas: schemas,
      result: this.create(result)
    };
  }
  /**
   * ObjectSchema.denormalize
   */
  ;

  _proto.denormalize = function denormalize(denormalizeState, path) {
    if (path === void 0) {
      path = [];
    }

    var result = denormalizeState.result,
        entities = denormalizeState.entities;
    var relations = this.relations;

    if (result == null || result === REMOVED_ENTITY) {
      return result;
    } // Map denormalize to the values of result, but only
    // if they have a corresponding schema. Otherwise return the plain value.
    // Then filter out deleted keys, keeping track of ones deleted
    // Then Pump the filtered object through `denormalizeFilter`


    var item = _extends({}, result);

    if (path.indexOf(this) !== -1) {
      return item;
    }

    for (var key in relations) {
      var schema = relations[key];
      var _result = item[key];
      var value = schema == null ? void 0 : schema.denormalize({
        result: _result,
        entities: entities
      }, [].concat(path, [this]));

      if (value !== REMOVED_ENTITY && value != undefined) {
        item[key] = value;
      } else {
        delete item[key];
      }
    }

    return item;
  };

  return ObjectSchema;
}();

export { ArraySchema, EntitySchema, ObjectSchema, REMOVED_ENTITY, UndefinedIdError };
//# sourceMappingURL=enty.esm.js.map
